/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package org.apache.hadoop.zebra.mapred;

import java.io.IOException;
import java.io.StringReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.BytesWritable;
import org.apache.hadoop.zebra.io.BasicTable;
import org.apache.hadoop.zebra.io.TableScanner;
import org.apache.hadoop.zebra.types.ParseException;
import org.apache.hadoop.zebra.types.Schema;

/**
 * Table expression for reading a BasicTable.
 * 
 * @see <a href="doc-files/examples/ReadABasicTable.java">Usage example for
 *      BasicTableExpr</a>
 */
class BasicTableExpr extends TableExpr {
  private Path path;

  /**
   * default constructor.
   */
  public BasicTableExpr() {
    // no-op
  }

  /**
   * Constructor.
   * 
   * @param path
   *          Path of the BasicTable.
   */
  public BasicTableExpr(Path path) {
    this.path = path;
  }

  /**
   * Set the path.
   * 
   * @param path
   *          path to the BasicTable.
   * @return self.
   */
  public BasicTableExpr setPath(Path path) {
    this.path = path;
    return this;
  }

  /**
   * Get the path.
   * 
   * @return the path to the BasicTable.
   */
  public Path getPath() {
    return path;
  }

  @Override
  protected BasicTableExpr decodeParam(StringReader in) throws IOException {
    String strPath = TableExprUtils.decodeString(in);
    if (strPath == null) {
      throw new RuntimeException("Incomplete expression");
    }
    path = new Path(strPath);
    return this;
  }

  @Override
  protected BasicTableExpr encodeParam(StringBuilder out) {
    if (path == null) {
      throw new RuntimeException("Incomplete expression");
    }
    TableExprUtils.encodeString(out, path.toString());
    return this;
  }

  @Override
  public List<LeafTableInfo> getLeafTables(String projection) {
    ArrayList<LeafTableInfo> ret = new ArrayList<LeafTableInfo>(1);
    ret.add(new LeafTableInfo(path, projection));
    return ret;
  }

  @Override
  public TableScanner getScanner(BytesWritable begin, BytesWritable end,
      String projection, Configuration conf) throws IOException {
    BasicTable.Reader reader = new BasicTable.Reader(path, conf);
    try {
      reader.setProjection(projection);
    } catch (ParseException e) {
    	throw new IOException("Projection parsing failed : "+e.getMessage());
    }
    return reader.getScanner(begin, end, true);
  }

  @Override
  public Schema getSchema(Configuration conf) throws IOException {
    return BasicTable.Reader.getSchema(path, conf);
  }

  @Override
  public boolean sortedSplitCapable() {
    return true;
  }
  
  @Override
  protected void dumpInfo(PrintStream ps, Configuration conf, int indent) throws IOException
  {
    BasicTable.dumpInfo(path.toString(), ps, conf, indent);
  }
}
